class NotRegistered(Exception):
    pass


class InvalidInput(Exception):
    pass


class TooRequests(Exception):
    pass


class InvaildAuth(Exception):
    pass


class Repeated(Exception):
    pass


from ..connections import Connections
from ...crypto import Crypto
from json import dumps, loads
from ...exceptions import (
    NotRegistered,
    InvaildAuth,
    InvalidInput,
    TooRequests,
    Repeated,
)



Web = {
        'app_name': 'Main',
        'app_version': '4.1.7',
        'platform': 'Web',
        'package': 'web.rubika.ir',
        'lang_code': 'fa'
}
Android = {
        'app_name': 'Main',
        'app_version': '3.0.7',
        'platform': 'Android',
        'package': 'ir.resaneh1.iptv',
        'lang_code': 'fa'
}


class MethodsMaker:
    __slots__ = (
        '__auth__',
        'cns',
        '__crypto__',
        '__uploader__',
    )
    def __init__(self, auth):
        self.__auth__ = auth
        self.cns = Connections(auth)
        self.__uploader__ = self.cns.uploadFile
        self.__crypto__ = Crypto(auth)

    async def request(self, method, data, method_type=5, custom_client=False):
        if method_type == 5:
            data = {
                'api_version': '5',
                'auth': self.__auth__,
                'data_enc': self.__crypto__.encode(dumps({
                    'method': method,
                    'input': data,
                    'client': Web if custom_client == False else Android
                })
            )}
            response = await self.cns.POST(data=data)
            response = loads(self.__crypto__.decode(response.get('data_enc')))
            status = response.get('status')
            if status == 'OK':
                return response.get('data')
            elif status == 'ERROR_GENERIC' or status == 'ERROR_ACTION':
                status_det = response.get('status_det')
                if status_det == 'NOT_REGISTERED':
                    raise NotRegistered('Your AUTH is incorrect')
                elif status_det == 'INVALID_INPUT':
                    raise InvalidInput('The data sent to the server is wrong')
                elif status_det == 'INVALID_AUTH':
                    raise InvaildAuth('An error was received from the server side, probably the data sent is wrong or your AUTH is invalid')
                elif status_det == 'TOO_REQUESTS':
                    raise TooRequests('Unfortunately, your account has been limited')
                else:
                    raise RuntimeWarning(response)

        elif method_type == 4:
            data = {
                'api_version': '4',
                'auth': self.__auth__,
                'client': Android,
                'method': method,
                'data_enc': self.__crypto__.encode(
                    dumps(data)
                )
            }
            response = await self.cns.POST(
                data=data,
            )
            status = response.get('status')
            if status == 'OK':
                return loads(self.__crypto__.decode(response.get('data_enc'))).get('data')
            elif status == 'ERROR_GENERIC' or status == 'ERROR_ACTION':
                status_det = response.get('status_det')
                if status_det == 'NOT_REGISTERED':
                    raise NotRegistered('Your AUTH is incorrect')
                elif status_det == 'INVALID_INPUT':
                    raise InvalidInput('The data sent to the server is wrong')
                elif status_det == 'INVALID_AUTH':
                    raise InvaildAuth('An error was received from the server side, probably the data sent is wrong or your AUTH is invalid')
                elif status_det == 'TOO_REQUESTS':
                    raise TooRequests('Unfortunately, your account has been limited')
                else:
                    raise RuntimeWarning(response)



from re import findall
try: from PIL import Image
except ModuleNotFoundError: Image = None
from io import BytesIO
from pybase64 import b64encode
from mutagen.mp3 import MP3
try: from tinytag import TinyTag
except ModuleNotFoundError: TinyTag = None


class Utils:
    __slots__ = ()
    def __init__(self):
        pass

    def textParser(self, text,):
        results = []
        real_text = text.replace('**', '').replace('__', '').replace('``', '')

        bolds = findall(r'\*\*(.*?)\*\*', text)
        italics = findall(r'\_\_(.*?)\_\_', text)
        monos = findall(r'\`\`(.*?)\`\`', text)
        mResult = [real_text.index(i) for i in monos]
        bResult = [real_text.index(i) for i in bolds]
        iResult = [real_text.index(i) for i in italics]

        for bIndex, bWord in zip(bResult, bolds):
            results.append({
				'from_index' : bIndex,
				'length' : len(bWord),
				'type' : 'Bold'
			})

        for iIndex, iWord in zip(iResult, italics):
            results.append({
				'from_index' : iIndex,
				'length' : len(iWord),
				'type' : 'Italic'
			})

        for mIndex, mWord in zip(mResult, monos):
            results.append({
				'from_index' : mIndex,
				'length' : len(mWord),
				'type' : 'Mono'
			})

        return (results, real_text)

    def adsFinder(self, string):
        urls = findall('http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+', string)
        if urls != []: return True
        elif '@' in string: return True
        for check in ['.ir', '.com', '.org']:
            if check in string: return True
        else:
            return False

    def is_forward(self, message):
        try:
            if 'forwarded_from' in message.keys() and message.get('forwarded_from').get('type_from') == 'Channel':
                return True
            else:
                return False
        except KeyError:
            message = message.get('message')
            if 'forwarded_from' in message.keys() and message.get('forwarded_from').get('type_from') == 'Channel':
                return True
            else:
                return False

    def getImageSize(self, image_bytes):
        if Image != None:
            image = Image.open(BytesIO(image_bytes))
            return image.size
        else:
            raise ImportWarning('Please install <pillow> and try again')

    def getThumbnail(self, image_bytes):
        if Image != None:
            image = Image.open(BytesIO(image_bytes))
            width, height = image.size
            if height > width:
                new_height = 40
                new_width  = round(new_height * width / height)
            else:
                new_width = 40
                new_height = round(new_width * height / width)
            image = image.resize((new_width, new_height), Image.ANTIALIAS)
            changed_image = BytesIO()
            image.save(changed_image, format='PNG')
            return b64encode(changed_image.getvalue())
        else:
            raise ImportWarning('Please install <pillow> and try again')

    async def get_voice_duration(self, file_bytes):
        file = BytesIO()
        file.write(file_bytes)
        file.seek(0)
        audio = MP3(file)
        return audio.info.length

    async def getVideoDuration(self, video):
        if TinyTag != None:
            duration = TinyTag.get(video).duration
            return round(duration * 1000)
        else:
            raise ImportWarning('Plaese install <TinyTag> and try again')

    async def getMusicArtist(self, music_bytes):
        return str(TinyTag.get(music_bytes).artist)

    async def getMimeFromURL(self, url):
        if '?' in url:
            return url.split('/')[-1].split('?')[0].split('.')[-1]
        elif '.' in url:
            return url.split('.')[-1]
        else:
            return '.unknown'

    async def thumb_inline(self):
        return 'iVBORw0KGgoAAAANSUhEUgAAACgAAAAoCAIAAAADnC86AAAAL0lEQVR4nO3NQQ0AAAgEIPVz/Rsbw81BATpJXZiTVSwWi8VisVgsFovFYrFY/DRelEIAZd5yXa4AAAAASUVORK5CYII='


from urllib3 import PoolManager
from websockets import connect, ConnectionClosed
from random import randint, choice
from json import dumps, loads
from time import time
from ...crypto import Crypto


__all__ = ('Connections', 'WebSocket')


class Connections:
    __slots__ = ('__manager__', '__auth__')

    def __init__(self, auth):
        self.__manager__ = PoolManager(headers={'Content-Type':'application/json'})
        self.__auth__ = auth

    async def GET(self, url, headers=None, download=False):
        async def runner():
            if download:
                while True:
                    response = self.__manager__.request('GET', url, headers=headers)
                    status = response.status
                    if status == 200:
                        return response.data
                    elif status == 502:
                        continue
                    else:
                        data = response.data.decode('UTF-8')
                        raise Exception(f'An error received from rubika server(download servers):\n\tError: {data or {}}\n\tstatus code: {status}')
            else:
                return loads(self.__manager__.request('GET', url, headers=headers).data.decode('UTF-8'))
        return await runner()

    async def getdcmess(self, ws=False,):
        url = 'https://getdcmess.iranlms.ir/'
        if ws:
            response = await self.GET(url=url)
            return response.get('data').get('socket').values()
        else:
            response = await self.GET(url=url)
            return response.get('data').get('API').values()

    async def POST(self, data):
        data = dumps(data)
        while True:
            response = self.__manager__.request('POST', url=f'https://messengerg2c{randint(1,2)}.iranlms.ir/', body=data.encode('UTF-8'))
            status = response.status
            if status == 200:
                return loads(response.data.decode('UTF-8'))
            elif status == 502:
                continue
            else:
                raise Exception('response status: {}\nresponse body: {}'.format(status, response.data.decode('UTF-8') or {}))

    async def uploadData(self, url, data, headers):
        while True:
            response = self.__manager__.request('POST', url=url, body=data, headers=headers)
            status = response.status
            if status == 200:
                return loads(response.data.decode('UTF-8'))
            elif status == 502:
                continue
            else:
                raise Exception('response status: {}\nresponse body: {}'.format(status, response.data.decode('UTF-8') or {}))

    async def uploadFile(self, upload_url, access_hash_send, file_id, file_bytes):
        size = str(len(file_bytes))
        headers = {
            'auth': self.__auth__,
            'chunk-size': size,
            'part-number': '1',
            'total-part': '1',
            'file-id': str(file_id),
            'access-hash-send': access_hash_send,
            'content-type': 'application/octet-stream',
        }

        if int(size) <= 131072:
            response = await self.uploadData(upload_url, file_bytes, headers)
            return response.get('data').get('access_hash_rec')

        else:
            size = int(size)
            total_part = size // 131072 + 1
            for part_number in range(1, total_part + 1):
                bsb = (part_number - 1) * 131072 # base set  file bytes
                headers["chunk-size"], headers["part-number"], headers["total-part"] = "131072" if part_number != total_part else str(len(file_bytes[bsb:])), str(part_number), str(total_part)
                if part_number != total_part:
                    data = file_bytes[bsb:bsb + 131072]
                    response = await self.uploadData(upload_url, data, headers)
                    assert response.get('data') == None
                else:
                    data = file_bytes[bsb:]
                    response = await self.uploadData(upload_url, data, headers)
                    return response.get('data').get('access_hash_rec')


class WebSocket:

    __slots__ = (
        '__auth__',
        '__crypto__',
        'getdcmess',
        'data',
    )

    def __init__(self, auth,):
        self.__auth__ = auth
        self.__crypto__ = Crypto(auth,)
        self.getdcmess = Connections(auth).getdcmess
        self.data = {
            'api_version': '5',
            'auth': auth,
            'method': 'handShake'
        }
        del auth

    async def handSnake(self,):
        wss = choice(list(await self.getdcmess(ws=True,),),)
        async for websocket in connect(wss,):
            try:
                await websocket.send(dumps(self.data,),)
                while True:
                    data = await websocket.recv()
                    if data != '{"status":"OK","status_det":"OK"}':
                        yield loads(data,)
                    else: continue
            except ConnectionClosed: continue

    async def updatesHandler(self, chat_updates=False, message_updates=True, show_notifications=False):
        async for message in self.handSnake():
            if message.get('type') == 'messenger':
                data = loads(self.__crypto__.decode(message.get('data_enc')))
                if message_updates and chat_updates and show_notifications:
                    yield message
                elif message_updates:
                    if 'message_updates' in data:
                        for i in data.get('message_updates'):
                            yield i
                elif chat_updates:
                    if 'chat_updates' in data:
                        for i in data.get('chat_updates'):
                            yield i
                elif show_notifications:
                    if 'show_notifications' in data:
                        for i in data.get('show_notifications'):
                            yield i